#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <inttypes.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/epoll.h>
#include <time.h>

#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <pwd.h>
#include <grp.h>

#include <utils/utils.h>

#include "proto.h"

/*
 * Local constants
 */
#define PORT			1234
#define LOOP_HEARTBEAT		1000
#define EPOLL_MAX_EVENTS	10

#define VERSION "0.0.1pre" BUILDDATE " (build " BUILDDATE " " BUILDTIME ")"

#define ascii2bcd(x)	({ (x > '0' && x < '9') ? x -= '0' : 0 })
#define strtobcd(x)	({ typeof(x) *__ptr = x; while( *__ptr ) *__ptr++ = ascii2bcd(*__ptr); })

#define log(level, msg, args...)	({ if( level <= globals.loglevel ) printf(msg, ## args); })

#define ts_diff(a, b) ({ \
		long long int __tmpA, __tmpB; \
		__tmpA = (a.tv_sec * 1000000000) + a.tv_nsec; \
		__tmpB = (b.tv_sec * 1000000000) + b.tv_nsec; \
		__tmpA - __tmpB; \
	})

enum {
	LOG_NONE = 0,
	LOG_CRITICAL,
	LOG_ERROR,
	LOG_WARNING,
	LOG_INFO,
	LOG_DEBUG,
};

/*
 * Trie values
 */
struct node {
	int cost;
	char gateway[MAX_GATEWAY_SIZE];

	/* current size: 54 bytes */
};

struct request {
	int type;
	int id;
	int version;
	int invalid;
	int status;
	long long duration;

	int keylen;
	char key[TRIE_MAX_DEPTH];
};

/*
 * Global vars
 */
static struct {
	FILE *log_fp;
	int running;
	int loglevel;

	struct trie_node trie;
} globals;

static struct {
	unsigned long requests_total;
	unsigned long requests_success;

} statistics;


/*
 * Generic routines
 */
static int decode_request( struct n_buffer *nbuf, struct request *req )
{
	if( !nbuf || !req )
		return -1;

	if( buffer_getnumleft(nbuf) < 1 ) {
		log( LOG_INFO, "Request too small\n" );
		return -1;
	}

	memset( req, 0, sizeof(struct request) );

	req->type = buffer_getu8( nbuf );

	while( buffer_getnumleft(nbuf) >= 2 ) {
		struct proto_type *pt;
		int8_t len = 0, type, id;

		id   = buffer_getu8( nbuf );
		type = buffer_getu8( nbuf );

		pt = lookup_type( id );
		if( !pt ) {					/* var ok? */
			log( LOG_INFO, "Unknown variable %d\n", id );
			return -1;
		}

		if( pt->type != type ) {			/* type ok? */
			log( LOG_INFO, "Variable %d type mismatch\n", pt->id );
			return -1;
		}

		if( !(pt->flags & FLAGS_REQUEST) ) {		/* variable allowed in request? */
			log( LOG_INFO, "Variable %d not allowed in request\n", pt->id );
			return -1;
		}

		if( pt->len ) {					/* length needed, ok and enough bytes left? */
			len = buffer_getu8( nbuf );

			if( pt->len < len ) {
				log( LOG_INFO, "Variable %d: Length %d in request exceeds limit %d\n", pt->id, len, pt->len );
				return -1;
			}

			if( buffer_getnumleft(nbuf) < len ) {
				log( LOG_INFO, "Variable %d: Not enough bytes left in buffer: %d, requested: %d\n",
						pt->id, buffer_getnumleft(nbuf), (int)len);
				return -1;
			}
		}

		switch( pt->id ) {
		case VAR_VERSION:
			req->version = buffer_getu32( nbuf );
			break;

		case VAR_LOOKUP:
			req->keylen = buffer_read( nbuf, req->key, len );
			break;

		case VAR_REQUESTID:
			req->id = buffer_getu32( nbuf );
			break;

		default:
			log( LOG_INFO, "Unhandled variable %d\n", pt->id );
			return -1;
		}

		log( LOG_DEBUG, "Variable %d successfully decoded (%d bytes left)\n", pt->id, buffer_getnumleft(nbuf) );
	}
	return 0;
}

static int encode_response( struct n_buffer *nbuf, struct request *req, char *value )
{
	char vars[10];
	int n = 0;

	if( !nbuf || !req )
		return -1;

	memset( vars, 0, sizeof(vars) );

	buffer_putu8( nbuf, req->type );

	if( req->type == REQ_TYPE_PONG )
		return 0;

	if( !req->status )
		return -1;

	/* status */
	vars[n++] = VAR_STATUS;

	/* request id */
	if( req->id )
		vars[n++] = VAR_REQUESTID;

	/* invalid parameter */
	if( req->status != 200 && req->invalid )
		vars[n++] = VAR_INVALID;

	/* successful request */
	if( req->status == 200 ) {
		/* gateway */
		if( !value )
			return -1;

		vars[n++] = VAR_GATEWAY;

		if( req->duration )
			vars[n++] = VAR_DURATION;
	}

	for( int x = 0; x < n; x++ ) {
		struct proto_type *pt;
		int8_t len = 0;

		pt = lookup_type( vars[x] );
		if( !pt )
			return -1;

		buffer_putu8( nbuf, pt->id );
		buffer_putu8( nbuf, pt->type );

		if( pt->id == VAR_GATEWAY )
			len = strlen( value );

		if( len )
			buffer_putu8( nbuf, len );
		else if( pt->len )
			buffer_putu8( nbuf, pt->len );

		switch( pt->id ) {
		case VAR_REQUESTID:
			buffer_putu32( nbuf, req->id );
			break;

		case VAR_STATUS:
			buffer_putu32( nbuf, req->status );
			break;

		case VAR_INVALID:
			buffer_putu32( nbuf, req->invalid );
			break;

		case VAR_GATEWAY:
			buffer_write( nbuf, value, strlen(value) );
			break;

		case VAR_DURATION:
			buffer_putu64( nbuf, req->duration );
			break;

		default:
			return -1;
		}
	}
	return 0;
}


/**
 * handle_request
 * \brief	Handle DHCP request
 * \param	fd	Socket with pending request
 * \return	0 on success, -1 on error
 */
static int handle_request( int fd )
{
	struct timespec ts_before, ts_after;
	struct sockaddr_in addr;
	socklen_t addrlen = sizeof(struct sockaddr_in);
	struct request req;
	struct n_buffer *nbuf;
	char *val = NULL;
	char buf[1024];
	int res, len = 0;

	nbuf = buffer_init( buf, sizeof(buf) );
	if( !nbuf ) {
		log( LOG_CRITICAL, "Failed to initialize network buffer\n" );
		return 0;
	}

	log( LOG_DEBUG, "Receiving request\n" );

	res = recvfrom( fd, buffer_getptr(nbuf), buffer_getsize(nbuf), 0, (struct sockaddr *)&addr, &addrlen );
	if( res <= 0 ) {
		perror( "receiving request" );
		return 0;
	}
	buffer_setlen( nbuf, res );

	log( LOG_DEBUG, "Decoding request\n" );

	res = decode_request( nbuf, &req );
	if( res < 0 ) {
		log( LOG_INFO, "Request is invalid\n" );

		req.status = 400;
		goto out;
	}

	if( req.type == REQ_TYPE_LOOKUP ) {

		if( req.version != PROTOCOL_VERSION ) {
			req.status  = 400;
			req.invalid = VAR_VERSION;
			goto out;
		}

		/*
		 * Decode request
		 */
		if( req.keylen > TRIE_MAX_DEPTH ) {
			log( LOG_INFO, "Request exceeds max limit\n" );

			req.status = 400;
			req.invalid = VAR_LOOKUP;
			goto out;
		}

		/* ascii to bcd */
#if 0
		strtobcd( buf );
#endif
		statistics.requests_total++;

		/*
		 * Lookup route
		 */
		clock_gettime( CLOCK_REALTIME, &ts_before );
		val = trie_get_partial( &globals.trie, req.key, req.keylen );
		clock_gettime( CLOCK_REALTIME, &ts_after );

		if( !val ) {
			/*
			 * No result
			 */
			log( LOG_INFO, "No route for key\n" );

			req.status = 404;
		}
		else {
			/*
			 * Encode response
			 */
			log( LOG_INFO, "Found route for key: %s\n", (char *)val );

			statistics.requests_success++;
			req.status   = 200;
			req.duration = ts_diff( ts_after, ts_before );
		}
	}
	else if( req.type == REQ_TYPE_PING ) {

		log( LOG_INFO, "Sending PONG\n" );

		req.type = REQ_TYPE_PONG;
	}

out:
	buffer_reset( nbuf );

	res = encode_response( nbuf, &req, (char *)val );
	if( res < 0 ) {
		log( LOG_ERROR, "Response invalid" );
		return -1;
	}

	log( LOG_DEBUG, "len: %d\n", len );

	res = sendto( fd, buffer_getptr(nbuf), buffer_getlen(nbuf), 0, (struct sockaddr *) &addr, addrlen );
	if( res < 0 )
		perror( "sending response" );

	return res;
}

/**
 * handle_sigint
 * \brief	SIGINT handler
 * \param[in]	signum	signal number
 */
static void handle_sigint( int signum )
{
	if( globals.running ) {
		fprintf( stderr, "Shutting down, please wait...\n" );
		globals.running = 0;
	}
}

/**
 * handle_sigterm
 * \brief	SIGTERM handler
 * \param[in]	signum	signal number
 */
static void handle_sigterm( int signum )
{
	if( globals.running ) {
		fprintf( stderr, "Shutting down, please wait...\n" );
		globals.running = 0;
	}
}


/**
 * print_usage
 * \brief	Print usage information on stdout
 */
static void print_usage( void )
{
	printf( "Usage: axrouted [-dhV] [-u <user> [-g <group>]]\n\n" );
	printf( "\t-V\t\tPrint version information\n" );
	printf( "\t-d\t\tRun as daemon\n" );
	printf( "\t-u <user>\tUser to run as\n" );
	printf( "\t-g <group>\tGroup to run as\n" );
	printf( "\t-p <port>\tPort to listen on\n" );
	printf( "\t-l <loglevel>\tLog level\n" );
	printf( "\t-h\t\tPrint this help screen\n" );
}

/**
 * print_version
 * \brief	Print version on stdout
 */
static void print_version( void ) {
	printf( "axrouted %s\n", VERSION );
}



/**
 * main
 * \brief	well...
 */
int main( int argc, char **argv )
{
	struct sockaddr_in	saddr;
	int sock, port = PORT;
#ifdef USE_IPV6
	struct sockaddr_in6	saddr6;
	int sock6;
#endif
	int daemonize = 0;
	int res, opt;

	uid_t runas_uid = 0;
	gid_t runas_gid = 0;
	struct passwd *runas_pw = NULL;

	int ep_fd;
	struct epoll_event ep_event;

	/*
	 * Defaults
	 */
	globals.loglevel = LOG_ERROR;

	/*
	 * options parsing
	 */
	while( ( opt = getopt(  argc, argv, "u:g:p:l:dhV" ) ) != -1 ) {
		struct group *gr;
		int val;

		switch( opt ) {
		case 'd':
			daemonize = 1;
			break;
		case 'p':
			port = atoi( optarg );
			if( port <= 0 || port > 65535 ) {
				printf( "Invalid port" );
				return -1;
			}
			break;
		case 'l':
			val = atoi( optarg );
			if( val < LOG_NONE ) {
				printf( "Invalid loglevel" );
				return -1;
			}
			if( val > LOG_DEBUG )
				val = LOG_DEBUG;

			globals.loglevel = val;
			break;
		case 'g':
			gr = getgrnam( optarg );
			if( !gr ) {
				perror( "gid" );
				return -1;
			}
			runas_gid = gr->gr_gid;
			break;
		case 'u':
			runas_pw = getpwnam( optarg );
			if( !runas_pw ) {
				perror( "uid" );
				return -1;
			}
			runas_uid = runas_pw->pw_uid;
			break;

		case 'V':
			print_version( );
			return 0;
		case 'h':
			print_usage( );
			return 0;
		}
	}

	/*
	 * daemonize
	 */
	if( daemonize ) {
		switch( fork() ) {
		case -1:
			/* error */
			perror( "fork" );
			return -1;
			break;
		case 0:
			/* child context */
			break;
		default:
			/* parent context */
			return 0;
		}
	}

	/*
	 *
	 */
	if( signal( SIGINT, &handle_sigint ) == SIG_ERR ) {
		perror( "Registering SIGINT callback" );
		return -1;
	}

	if( signal( SIGTERM, &handle_sigterm ) == SIG_ERR ) {
		perror( "Registering SIGTERM callback" );
		return -1;
	}

	/*
	 * create sockets
	 */
	sock = socket( PF_INET, SOCK_DGRAM, 0 );
	if( sock < 0 ) {
		perror( "Creating IPv4 socket" );
		return -1;
	}

#ifdef USE_IPV6
	sock6 = socket( PF_INET6, SOCK_DGRAM, 0 );
	if( sock6 < 0 ) {
		perror( "Creating IPv6 socket" );
		return -1;
	}
#endif

	/*
	 * bind sockets
	 */
	saddr.sin_addr.s_addr = INADDR_ANY;
	saddr.sin_port = htons( port );

	res = bind( sock, (struct sockaddr *)&saddr, sizeof(saddr) );
	if( res < 0 ) {
		perror( "Binding to IPv4 address:port" );
		return -1;
	}

#ifdef USE_IPV6
	saddr6.sin6_addr = in6addr_any;
	saddr6.sin6_port = htons( port );

	res = bind( sock6, (struct sockaddr *)&saddr6, sizeof(saddr6) );
	if( res < 0 ) {
		perror( "Binding to IPv6 address:port" );
		return -1;
	}
#endif

	/*
	 * drop privileges
	 */
	if( runas_uid ) {
		res = setgroups( 0, NULL );
		if( res < 0 ) {
			perror( "setgroups" );
			return -1;
		}

		if( runas_gid ) {
			res = setgid( runas_gid );
			if( res < 0 ) {
				perror( "setgid" );
				return -1;
			}
		} else {
			res = initgroups( runas_pw->pw_name, runas_pw->pw_gid );
			if( res < 0 ) {
				perror( "initgroups" );
				return -1;
			}
		}

		res = setuid( runas_uid );
		if( res < 0 ) {
			perror( "setuid" );
			return -1;
		}
	}

	/*
	 * Open logfile, close std{in,out,err}
	 */
	if( daemonize ) {
		fclose( stdin );

		globals.log_fp = freopen( LOGFILE, "a", stdout );
		if( !globals.log_fp ) {
			perror( "open logfile" );
			return -1;
		}

		fclose( stderr );

		stderr = globals.log_fp;

		/*
		 * Disable buffering 
		 */
		res = setvbuf( globals.log_fp, NULL, _IONBF, 0 );
		if( res < 0 )
			perror( "disabling output buffer" );
	}

	/*
	 * write pid file
	 */
	if( daemonize ) {
		FILE *fp;

		fp = fopen( PIDFILE, "w" );
		if( !fp ) {
			perror( "opening pid file" );
			return -1;
		}

		fprintf( fp, "%d", getpid() );
		fclose( fp );
	}

	/*
	 * setup epoll()
	 */
	ep_fd = epoll_create( 10 );
	if( ep_fd < 0 ) {
		perror( "creating epoll context" );
		return -1;
	}

	ep_event.events = EPOLLIN;
	ep_event.data.fd = sock;

	res = epoll_ctl( ep_fd, EPOLL_CTL_ADD, sock, &ep_event );
	if( res < 0 ) {
		perror( "adding IPv4 socket to epoll context" );
		return -1;
	}

#ifdef USE_IPV6
	ep_event.data.fd = sock6;

	res = epoll_ctl( ep_fd, EPOLL_CTL_ADD, sock6, &ep_event );
	if( res < 0 ) {
		perror( "adding IPv6 socket to epoll context" );
		return -1;
	}
#endif
	/*
	 * Init trie + populate
	 */
	trie_init( &globals.trie );

	{
		FILE *fp = NULL;
		char tmp[1024], *key, *value;
		int routes = 0, linenr = 0;

		fp = fopen( "/etc/axrouted/route.list", "r" );
		if( !fp ) {
			perror( "Failed to open routes file" );
			goto cleanup;
		}

		while( !feof( fp ) ) {
			linenr++;

			memset( tmp, 0, sizeof(tmp) );

			res = fread( tmp, sizeof(tmp), 1, fp );
			if( res < 0 ) {
				printf( "%d: error on read\n", linenr );
				break;
			}

			res = strnlen( tmp, sizeof(tmp) - 1 );
			if( !res ) {
				printf( "%d: empty line\n", linenr );
				continue;
			}

			tmp[res] = '\0';
			key   = tmp;
			value = tmp + (res - 1);

			while( *key && *key == ' ' ) key++;	/* ltrim() */
			while( *value == '\r' || *value == '\n' ) *value-- = '\0';	/* strip newline */

			value = strchr( key, ' ' );
			if( !value ) {
				value = strchr( key, '\t' );
				if( !value ) {
					printf( "%d: key-value delimiter not found\n", linenr );
					continue;
				}
			}

			*value++ = '\0';

			if( !strlen(value) ) {
				printf( "%d: value too short\n", linenr );
				continue;
			}

			res = trie_add( &globals.trie, key, strlen(key), strdup(value) );
			if( res < 0 )
				continue;

			routes++;
		}

		printf( "%d routes added\n", routes );

		fclose( fp );
	}


	/*
	 * main loop
	 */
	globals.running = 1;

	printf( "axrouted version %s up and running\n", VERSION );

	while( globals.running ) {
		struct epoll_event ep_events[EPOLL_MAX_EVENTS];
		int nfds;

		nfds = epoll_wait( ep_fd, ep_events, EPOLL_MAX_EVENTS, LOOP_HEARTBEAT );
		if( nfds < 0 ) {
			/* error condition */
			switch( errno ) {
			default:
				perror( "waiting for I/O" );
				globals.running = 0;
			case EINTR:
				continue;
			}
		}
		else if( nfds > 0 ) {

			/* data available */
			log( LOG_DEBUG, "data available on %d sockets\n", nfds );

			for( int n=0; n<nfds; n++ ) {
				if( ep_events[n].events & EPOLLIN ) {
					res = handle_request( ep_events[n].data.fd );
				} else
					printf( " unknown event\n" );
			}
		}
		else {
			/* timeout */
		}
	}

	/*
	 * cleanup
	 */
cleanup:
	trie_destroy( &globals.trie );

	close( ep_fd );
	close( sock );
#ifdef USE_IPV6
	close( sock6 );
#endif

	/*
	 * Remove pidfile, close logfile
	 */
	if( daemonize ) {
		res = unlink( PIDFILE );
		if( res < 0 )
			perror( "removing pidfile" );

	}
	if( globals.log_fp ) {
		res = fclose( globals.log_fp );
		if( res < 0 )
			perror( "closing logfile" );
	}

	printf( "Statistics: %lu Requests successfully served, %lu Total\n",
		statistics.requests_success, statistics.requests_total );

	return 0;
}
