
#include "common.h"
#include "buffer.h"


/*
 * Buffer
 */
#define CANARY_VAL 0xCCAACCDD
#define CHECK_CANARY(x) 					\
	({							\
		int32_t __tmp = CANARY_VAL;			\
		if( memcmp( x, &__tmp, sizeof(int32_t) ) ) {	\
			fprintf( stderr, "ALERT: Buffer canary damaged: %x != %x\n", *x, CANARY_VAL );	\
			exit( 1 );				\
		}						\
	})

struct n_buffer *buffer_init( char *buf, int size )
{
	struct n_buffer *tmp = (struct n_buffer *)buf;
	int32_t canary = CANARY_VAL;

	if( !tmp )
		return NULL;

	memset( tmp, 0, size );

	tmp->size = size - (sizeof(struct n_buffer) + sizeof(int32_t));

	if( tmp->size <= 0 )
		return NULL;

	memcpy( &tmp->raw[tmp->size], &canary, sizeof(int32_t) );

	return tmp;
}


struct n_buffer *buffer_create( int size )
{
	struct n_buffer *buf;
	int32_t canary = CANARY_VAL;

	buf = malloc( sizeof(struct n_buffer) + size + sizeof(int32_t));
	if( !buf )
		return NULL;

	memset( buf, 0, sizeof(struct n_buffer) + size );

	buf->size = size;

	memcpy( &buf->raw[size], &canary, sizeof(int32_t) );

	return buf;
}

void buffer_destroy( struct n_buffer *buf )
{
	CHECK_CANARY( &buf->raw[buf->size] );

	free( buf );
}

void buffer_reset( struct n_buffer *buf )
{
	buf->cur_pos = 0;
	buf->end_pos = 0;

	memset( buf->raw, 0, buf->size );
}

long buffer_getpos( const struct n_buffer *buf )
{
	return buf->cur_pos;
}

void buffer_setpos( struct n_buffer *buf, const long pos )
{
	buf->cur_pos = ( pos > buf->end_pos ) ? buf->end_pos : pos;
}

void buffer_seek( struct n_buffer *buf, const buffer_seek_t type, const long pos )
{
	switch( type ) {
	case BUFFER_SEEK_SET:
		buf->cur_pos = pos;

		if( buf->cur_pos < 0 )
			buf->cur_pos = 0;
		else if( buf->cur_pos >= buf->size )
			buf->cur_pos = buf->size - 1;

		break;

	case BUFFER_SEEK_CUR:
		buf->cur_pos += pos;

		if( buf->cur_pos < 0 )
			buf->cur_pos = 0;
		else if( buf->cur_pos >= buf->size )
			buf->cur_pos = buf->size - 1;
		break;

	case BUFFER_SEEK_END:
		buf->cur_pos = buf->end_pos + pos;

		if( buf->cur_pos < 0 )
			buf->cur_pos = 0;
		else if( buf->cur_pos >= buf->size )
			buf->cur_pos = buf->size - 1;
		break;
	}

	/* update end pointer for write seek */
	if( buf->cur_pos > buf->end_pos )
		buf->end_pos = buf->cur_pos;
}

/*
 * Low-level direct content access
 */
long buffer_getsize( const struct n_buffer *buf )
{
	return buf->size;
}

char * buffer_getptr( const struct n_buffer *buf )
{
	return (char *)buf->raw;
}

char * buffer_getcurrent( const struct n_buffer *buf )
{
	return (char *)&buf->raw[buf->cur_pos];
}

long buffer_getlen( const struct n_buffer *buf )
{
	return buf->end_pos;
}

void buffer_setlen( struct n_buffer *buf, const long len )
{
	buf->end_pos = ( len >= buf->size ) ? buf->size - 1 : len;
}

long buffer_getnumleft( const struct n_buffer *buf )
{
	return ( buf->end_pos - buf->cur_pos );
}

long buffer_getnumfree( const struct n_buffer *buf )
{
	return ( buf->size - buf->end_pos );
}


/*
 * Higher-level access
 */
size_t buffer_read( struct n_buffer *buf, void *ptr, size_t num )
{
	CHECK_CANARY( &buf->raw[buf->size] );

	if( num > ( buf->end_pos - buf->cur_pos ) )
		num = buf->end_pos - buf->cur_pos;

	if( num <= 0 )
		return num;

	memcpy( ptr, &buf->raw[buf->cur_pos], num );
	buf->cur_pos += num;

	return num;
}

size_t buffer_write( struct n_buffer *buf, const void *ptr, size_t num )
{
	CHECK_CANARY( &buf->raw[buf->size] );

	if( num > ( buf->size - buf->cur_pos ) )
		num = buf->size - buf->cur_pos;

	if( num <= 0 )
		return num;

	memcpy( &buf->raw[buf->cur_pos], ptr, num );
	buf->cur_pos += num;

	/* update end pointer if neccessary */
	if( buf->cur_pos > buf->end_pos )
		buf->end_pos = buf->cur_pos;

	return num;
}

int buffer_get( struct n_buffer *buf, void *var, const size_t size )
{
	if( buf->cur_pos + size > buf->end_pos )
		return -1;

	buffer_read( buf, var, size );
	return 0;
}

int buffer_put( struct n_buffer *buf, const void *var, const size_t size )
{
	if( buf->cur_pos + size >= buf->size )
		return -1;

	buffer_write( buf, var, size );
	return 0;
}


int buffer_memcmp( struct n_buffer *buf, const void *val, const size_t size )
{
	int res;

	CHECK_CANARY( &buf->raw[buf->size] );

	if( buf->cur_pos + size >= buf->size )
		return -1;

	res = memcmp( &buf->raw[buf->cur_pos], val, size );
	buf->cur_pos += size;

	return res;
}

/*
 * debug
 */
void buffer_print( struct n_buffer *buf )
{
	CHECK_CANARY( &buf->raw[buf->size] );

	printf( "\tbuffer (%s:%d):\n\t\t size: %ld, len: %ld, curpos: %ld\n", __FILE__, __LINE__, buffer_getsize( buf ), buffer_getlen( buf ), buffer_getpos( buf ) );

	#define _COLS	16

	for(int x=0,y=1; x < buffer_getlen( buf ); x++,y++) {
		int newline = ( x % _COLS == 0 );
		int last_in_row = ( x+1 >= buffer_getlen( buf ) || y % _COLS == 0 );

		if( newline )
			printf( "\n\t%03d: ", x );

		printf( "%02hhx", buffer_getptr( buf )[x] );

		if( buffer_getpos( buf ) == x )
			putchar( '*' );
		else if( !last_in_row )
			putchar( ':' );

		if( last_in_row && ( buffer_getpos( buf ) < x && buffer_getpos( buf ) + _COLS > x ) )
			printf( " <- curpos: %ld", buffer_getpos( buf ) );
	}
	printf( "\n" );
}
