#ifndef _COROUTINE_H
#define _COROUTINE_H

/*
 *
 */
#if defined(__hppa__)
#include <signal.h>
#include <setjmp.h>
#else
#include <ucontext.h>
#define USE_SWAPCONTEXT
#endif

/**
 * coro_state
 * \brief	coroutine task states
 */
typedef enum {
	CORO_INIT = 0,
	CORO_RUNNING,
	CORO_PAUSED,
	CORO_SHUTDOWN,
	CORO_FINISHED
} coro_state;

/** 
 * coro_task
 * \brief	a coroutine task
 */
struct coro_task {
	struct dlist_head 	list;		/*!< task list handle */

	int 			id;		/*!< task id */

#ifdef USE_SWAPCONTEXT
	ucontext_t	 	env;		/*!< user context */
#else
	sigjmp_buf		env;
#endif
	coro_state 		state;		/*!< state */
	struct coro_scheduler	*sched;		/*!< scheduler context */

	char 			stack[SIGSTKSZ];/*!< stack */

	void (* func)( struct coro_task *tsk );	/*!< function running in this task */
};


/**
 * coro_scheduler
 * \brief	coroutine scheduler context
 */
struct coro_scheduler {
	struct dlist_head tasks;	/*!< list of tasks controlled by this scheduler */

#ifdef USE_SWAPCONTEXT
	ucontext_t 	env;		/*!< user context */
#else
	sigjmp_buf	env;
#endif
	int 		abort;		/*!< abort after context switch */
	int 		last_id;	/*!< id of the last task */
	int 		active_tasks;	/*!< number of tasks (still) running */
};


/**
 * coro_init
 * \brief	Initialize coroutine context
 * \param	sched	pointer to a coro_sched struct
 * \return	0 on success, -1 on error
 */
int coro_init( struct coro_scheduler *sched );


/**
 * coro_add_task
 * \brief	Add task to coroutine context
 * \param	sched	pointer to a coro_sched struct
 * \param	func	pointer to coroutine function
 * \return	0 on success, -1 on error
 */
int coro_add_task( struct coro_scheduler *sched, void (* func)( struct coro_task *tsk ) );


/**
 * coro_getid
 * \brief	Get coroutine id
 * \param	tsk	pointer to a coro_task struct
 * \return	0 on success, -1 on error
 * \note	propably only useful in a coroutine
 */
inline int coro_getid( const struct coro_task *tsk );


/**
 * coro_main
 * \brief	Coroutine context main loop
 * \param	sched	pointer to a coro_sched struct
 * \return	0 on success, -1 on error
 */
int coro_main( struct coro_scheduler *sched );


/**
 * coro_yield
 * \brief	yield to next coroutine
 * \param	tsk	pointer to current task struct
 * \return	0 on success, -1 on error
 * \note	call only from within coroutine!
 */
inline int coro_yield( struct coro_task *tsk );


/**
 * coro_return
 * \brief	return from coroutine
 * \param	tsk	pointer to current task struct
 * \note	call only from within coroutine!
 */
inline void coro_return( struct coro_task *tsk );


/**
 * coro_find
 * \brief	find a task by its id
 * \param[in]	tsk	current task
 * \param[in]	id	task id to search for
 * \return	pointer to the struct coro_task on success, NULL on error
 * \note	call from within coroutine
 */
struct coro_task * coro_find( const struct coro_task *tsk, const int id );


/**
 * coro_yieldto
 * \brief	yield to another task, skipping the scheduler
 * \param[in]	tsk	current task
 * \param[in]	to	task to switch to
 * \return	0 success (on the next timeslice), -1 on error
 * \note	call only from within coroutine!
 */
inline int coro_yieldto( struct coro_task *tsk, struct coro_task *to );


#endif
