#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/time.h>
#include <time.h>
#include <errno.h>

#include "bitmap.h"
#include "trie.h"

#define tv_diff(a, b) ({ \
		long long int __tmpA, __tmpB; \
		__tmpA = (a.tv_sec * 1000000) + a.tv_usec; \
		__tmpB = (b.tv_sec * 1000000) + b.tv_usec; \
		__tmpA - __tmpB; \
	})

#define ts_diff(a, b) ({ \
		long long int __tmpA, __tmpB; \
		__tmpA = (a.tv_sec * 1000000000) + a.tv_nsec; \
		__tmpB = (b.tv_sec * 1000000000) + b.tv_nsec; \
		__tmpA - __tmpB; \
	})

//#define BENCHMARKS
#define NUM_ITER 1000000

int main( void )
{
	int num_tests = 0, num_failed = 0, failed = 0;
	float tdiff = 0.0f;

//	struct trie_head trie;
	struct trie_node trie;
	int val1 = 12345;
	int val2 = 54321;
	int *ptr = NULL;
	const char *key1 = "tea";
	const char *key2 = "tealeaf";
	const char *key3 = "coffee";
	const char *key4 = "teabreak";

	printf( "\n\nTests:\n" );

	trie_init( &trie );

	/*
	 * Test 1: Add key
	 */
	num_tests++;
	printf( "\n*** Test 1: Add key\n" );

	printf( "   Adding key \"tea\" with ptr %p to val %d\n", &val1, val1 );
	trie_add( &trie, key1, strlen(key1), &val1 );

	printf( "   Key \"tea\" in trie? %s\n", trie_has_key( &trie, key1, strlen(key1) ) ? "Yes" : "No" );

	ptr = trie_get( &trie, key1, strlen(key1) );
	printf( "   Value for key \"tea\" = %d\n", (ptr) ? *ptr : 0 );

	if( !ptr || *ptr != val1 ) {
		printf( "!!! Test failed: %d != %d\n", (ptr) ? *ptr : 0, val1 );
		num_failed++;
	} else
		printf( "OK: Key successfully added and retrieved\n" );


	/*
	 * Test 2: Add another key with overlapping prefix
	 */
	num_tests++;
	printf( "\n*** Test 2: Add key with overlapping prefix\n" );

	printf( "   Adding key \"tealeaf\" with ptr %p to val %d\n", &val2, val2 );
	trie_add( &trie, key2, strlen(key2), &val2 );

	printf( "   Key \"tealeaf\" in trie? %s\n", trie_has_key( &trie, key2, strlen(key2) ) ? "Yes" : "No" );

	ptr = trie_get( &trie, key2, strlen(key2) );
	printf( "   Value for key \"tealeaf\" = %d\n", (ptr) ? *ptr : 0 );

	if( *ptr != val2 ) {
		printf( "!!! Test failed: %d != %d\n", *ptr, val2 );
		num_failed++;
	} else
		printf( "OK: Key with shared prefix successfully added and retrieved\n" );

	/*
	 * Test 2b: Check old key value
	 */
	num_tests++;
	printf( "\n*** Test 2b: First key value still OK?\n" );

	ptr = trie_get( &trie, key1, strlen(key1) );
	printf( "   Value for key \"tea\" = %d\n", (ptr) ? *ptr : 0 );

	if( *ptr != val1 ) {
		printf( "!!! Test failed: %d != %d\n", *ptr, val1 );
		num_failed++;
	} else
		printf( "OK: Key added without affecting already existing one\n" );


	/*
	 * Test 3: Non-existent key
	 */
	num_tests++;
	printf( "\n*** Test 3: Non-existent key\n" );

	printf( "   Key \"coffee\" in trie? %s\n", trie_has_key( &trie, key3, strlen(key3) ) ? "Yes" : "No" );

	ptr = trie_get( &trie, key3, strlen(key3) );
	if( ptr )
		printf( "   Value for key \"coffee\" = %d\n", *ptr );

	if( ptr != NULL ) {
		printf( "!!! Test failed: %p != NULL\n", ptr );
		num_failed++;
	} else
		printf( "OK: Key does not exist\n" );


	/*
	 * Test 3b: Non-existant key partially overlapping with existing one
	 */
	num_tests++;
	printf( "\n*** Test 3b: Non-existent key partially overlapping with existing one\n" );

	printf( "   Key \"teabreak\" in trie? %s\n", trie_has_key_partial( &trie, key4, strlen(key4) ) ? "Yes" : "No" );

	ptr = trie_get_partial( &trie, key4, strlen(key4) );
	if( ptr ) {
		printf( "   Value for key \"teabreak\" = %d\n", *ptr );
		printf( "OK: Partial matchin Key does exist: %p\n", ptr );
	}
	else {
		printf( "!!! Test failed: ptr = NULL\n" );
		num_failed++;
	}

	/*
	 * Test 4: Removing key
	 */
	num_tests++;
	printf( "\n*** Test 4: Removing key\n" );

	printf( "   Removing key \"tealeaf\"\n" );

	trie_del( &trie, key2, strlen(key2) );

	printf( "   Key \"tealeaf\" still in trie? %s\n", trie_has_key( &trie, key2, strlen(key2) ) ? "Yes" : "No" );

	printf( "   Key \"tea\" still in trie? %s\n", trie_has_key( &trie, key1, strlen(key1) ) ? "Yes" : "No" );

	failed = 0;
	if( !trie_get( &trie, key1, strlen(key1) ) ) {
		printf( "!!! Test failed: %d != %d\n", *ptr, val1 );
		failed++;
	}

	if( trie_get( &trie, key2, strlen(key2) ) ) {
		printf( "!!! Test failed: %d != %d\n", *ptr, val1 );
		failed++;
	}

	if( failed )
		num_failed++;
	else
		printf( "OK: Removed key without affecting other one sharing the same prefix\n" );

	/*
	 * Test 4b: Removing short key sharing same prefix with a longer one
	 */
	num_tests++;
	printf( "\n*** Test 4b: Removing short key sharing same prefix with longer one\n" );

	printf( "   Readding key \"tealeaf\"\n" );

	trie_add( &trie, key2, strlen(key2), &val2 );

	printf( "   Key \"tealeaf\" in trie? %s\n", trie_has_key( &trie, key2, strlen(key2) ) ? "Yes" : "No" );


	printf( "   Removing key \"tea\"\n" );

	trie_del( &trie, key1, strlen(key1) );

	printf( "   Key \"tea\" still in trie? %s\n", trie_has_key( &trie, key1, strlen(key1) ) ? "Yes" : "No" );

	printf( "   Key \"tealeaf\" still in trie? %s\n", trie_has_key( &trie, key2, strlen(key2) ) ? "Yes" : "No" );

	failed = 0;
	if( trie_get( &trie, key1, strlen(key1) ) ) {
		printf( "!!! Test failed: %d != %d\n", *ptr, val1 );
		failed++;
	}

	if( !trie_get( &trie, key2, strlen(key2) ) ) {
		printf( "!!! Test failed: %d != %d\n", *ptr, val1 );
		failed++;
	}

	if( failed )
		num_failed++;
	else
		printf( "OK: Removed key without affecting other one sharing the same prefix\n" );


	/*
	 * Test 5: Multiple colliding keys
	 */
	num_tests++;
	printf( "\n*** Test 5: Multiple colliding keys\n" );

	{
		const char mkey1[] = { 'c', 'o', 'f', 'f', 'e', '\0' };
		const char mkey2[] = { 'c', 'o', 'f', 'f', 'e', '\0', '\0' };

		failed = 0;

		printf( "    Adding mkey1 \"coffee\", val %d\n", val1 );
		trie_add( &trie, mkey1, sizeof(mkey1)-1, &val1 );

		printf( "    Adding mkey2 \"coffee0\", val %d\n", val2 );
		trie_add( &trie, mkey2, sizeof(mkey2)-1, &val2 );

		ptr = trie_get( &trie, mkey1, sizeof(mkey1)-1 );
		if( ptr && *ptr == val1 )
			printf( "   Key mkey1 value = %d\n", *ptr );
		else {
			failed++;
			if( !ptr )
				printf( "!!! Key mkey1 value is NULL\n" );
			else
				printf( "!!! Key mkey1 value is wrong: %d expected %d\n", *ptr, val1 );
		}

		ptr = trie_get( &trie, mkey2, sizeof(mkey2)-1 );
		if( ptr && *ptr == val2 )
			printf( "   Key mkey2 value = %d\n", *ptr );
		else {
			failed++;
			if( !ptr )
				printf( "!!! Key mkey2 value is NULL\n" );
			else
				printf( "!!! Key mkey2 value is wrong: %d expected %d\n", *ptr, val2 );
		}

		if( failed )
			num_failed++;
		else
			printf( "OK: trie_add() works as expected\n" );
	}

	/*
	 * Test 6: Adding an existing key
	 */
	num_tests++;
	printf( "\n*** Test 6: Adding already existing key\n" );

	{
		const char mkey1[] = { 'c', 'o', 'f', 'f', 'e', '\0' };

		if( trie_add( &trie, mkey1, sizeof(mkey1) - 1, &val1 ) < 0 )
			printf( "OK: Failed to add key (%s)\n", strerror(errno) );
		else {
			num_failed++;
			printf( "!!! Shouldn't be abled to add an already existing key with trie_add()\n" );
		}
	}

#ifdef BENCHMARKS
	printf( "\n\nBenchmarks:\n" );

	/*
	 * Prepare trie (readd keys)
	 */
	trie_add( &trie, key1, strlen(key1), &val1 );
	trie_add( &trie, key2, strlen(key2), &val2 );

	/*
	 * Test 10: Benchmark, non-recursive trie_get
	 */
	num_tests++;
	printf( "\n*** Test 10: Benchmark of trie_get\n" );

	{
		struct timeval tv_before, tv_after;

		// non-recursive
		gettimeofday( &tv_before, NULL );
		for( int i=0; i<NUM_ITER; i++ ) {
			ptr = trie_get( &trie, key2, strlen(key2) );
		}
		gettimeofday( &tv_after, NULL );

		tdiff = (float)tv_diff( tv_after, tv_before );
		printf( "   %d rounds of iterative trie_get = %-5.3f msecs (%f usecs per round)\n", NUM_ITER, tdiff / 1000, tdiff / NUM_ITER );

		if( ptr && *ptr )
			printf( "   retrieved value = %d\n", *ptr );
	}


	/*
	 * Test 10b: Benchmark, non-recursive trie_get_partial
	 */
	num_tests++;
	printf( "\n*** Test 10b: Benchmark of trie_get_partial\n" );

	{
		struct timeval tv_before, tv_after;

		// non-recursive
		gettimeofday( &tv_before, NULL );
		for( int i=0; i<NUM_ITER; i++ ) {
			ptr = trie_get_partial( &trie, key2, strlen(key2) );
		}
		gettimeofday( &tv_after, NULL );

		tdiff = (float)tv_diff( tv_after, tv_before );
		printf( "   %d rounds of iterative trie_get_partial = %-5.3f msecs (%f usecs per round)\n", NUM_ITER, tdiff / 1000, tdiff / NUM_ITER );

		if( ptr && *ptr )
			printf( "   retrieved value = %d\n", *ptr );
	}

	/*
	 * Test 11: Benchmark, trie_add + trie_del
	 */
	num_tests++;
	printf( "\n*** Test 11: Benchmark of trie_add + trie_del\n" );

	{
		struct timeval tv_before, tv_after;

		// recursive
		gettimeofday( &tv_before, NULL );
		for( int i=0; i<NUM_ITER; i++ ) {
			trie_add( &trie, key2, strlen(key2), &val1 );
			trie_del( &trie, key2, strlen(key2) );
		}
		gettimeofday( &tv_after, NULL );

		tdiff = (float)tv_diff( tv_after, tv_before );
		printf( "   %d rounds of trie_add + trie_del       = %-5.3f msecs (%f usecs per round)\n", NUM_ITER, tdiff / 1000, tdiff / NUM_ITER );
	}


	/*
	 * Test 12: Benchmark, single trie_add, trie_del
	 */
	num_tests++;
	printf( "\n*** Test 12: Benchmark of trie_add + trie_del\n" );

	{
		struct timespec ts_before, ts_after;

		// recursive
		clock_gettime( CLOCK_REALTIME, &ts_before );
		trie_add( &trie, key2, strlen(key2), &val1 );
		clock_gettime( CLOCK_REALTIME, &ts_after );

		tdiff = (float)ts_diff( ts_after, ts_before ) / 1000;
		printf( "   1 round of trie_add       = %-5.3f usecs\n", tdiff );


		clock_gettime( CLOCK_REALTIME, &ts_before );
		trie_del( &trie, key2, strlen(key2) );
		clock_gettime( CLOCK_REALTIME, &ts_after );

		tdiff = (float)ts_diff( ts_after, ts_before ) / 1000;
		printf( "   1 round of trie_del       = %-5.3f usecs\n", tdiff );
	}
#endif
	printf( "\n*** Done, destroying trie\n" );
	trie_destroy( &trie );


	/*
	 * Test 13: VoIP related longest prefix matching benchmark + tests
	 */
	num_tests++;
	printf( "\n*** Test 13: VoIP related longest prefix matching benchmark + tests\n" );

	{
		int x=0, sum_length=0, num_routes=0, failed_routes=0;
		char *val = NULL, *key = NULL;
		struct timespec ts_before, ts_after;
		struct trie_node lpm_trie;
		struct {
			char prefix[25];
		} prefixes[] = {
			/* DE */
			{ "0049" },
			{ "0049221" },
			{ "00492219139170" },
			{ "00492238" },

			/* U.S.*/
			{ "001" },
			{ "001500" },
		};
		/* init new trie */

		trie_init( &lpm_trie );

		for(x=0; x<(sizeof(prefixes) / sizeof(*prefixes)); x++) {
			printf( "   adding route prefix %-25s\n", prefixes[x].prefix );
			trie_add( &lpm_trie, prefixes[x].prefix, strlen(prefixes[x].prefix), prefixes[x].prefix );
		}
		printf( "   %d prefixes added\n", x );

		/* lookup tests */
		printf( "\n  * Part 1: LPM Route lookup tests\n" );

		#define lpm_test1(_key) ({							\
			char *_x = _key;							\
			val = trie_get_partial( &lpm_trie, _x, strlen(_x) );			\
			if( val )								\
				printf( "   key %-25s -> prefix %s\n", _x, val );		\
			else									\
				printf( "   failed to find route for key %-25s\n", _x );	\
			num_routes++;								\
			sum_length+= strlen(_x);						\
		})
			

		clock_gettime( CLOCK_REALTIME, &ts_before );
		for(x=0; x<10; x++) {
			lpm_test1( "004922191391721" );	/* DE: 0049221 prefix */
			lpm_test1( "00492238922733" );	/* DE: 00492238 prefix */
			lpm_test1( "004922251234567" );	/* DE: 0049 generic prefix */
			lpm_test1( "0015234" );		/* US: 001 prefix */
			lpm_test1( "00150089" );	/* US: 001500 prefix */
		}
		clock_gettime( CLOCK_REALTIME, &ts_after );

		tdiff = (float)ts_diff( ts_after, ts_before ) / 1000;

		/* output some stats */
		printf( "\n   --- Stats ---\n" );
		printf( "   Number of lookups performed: %d keys\n", num_routes); 
		printf( "   Average route key length   : %.2f chars\n", ((float)sum_length/(float)num_routes));
		printf( "   Time for lookups           : %-5.3f usecs\n", tdiff );
		printf( "   Average time per lookup    : %-5.3f usecs\n", (tdiff/(float)num_routes) );


		/*
		 * 13.2: speed
		 */
		printf( "\n  * Part 2: LPM Route lookup tests (no output on successfull match)\n" );
		num_routes=0;
		sum_length=0;

		#define lpm_test1_quiet(_key) ({							\
			char *_x = _key;							\
			val = trie_get_partial( &lpm_trie, _x, strlen(_x) );			\
			if( !val )								\
				printf( "   failed to find route for key %-25s\n", _x );	\
			num_routes++;								\
			sum_length+= strlen(_x);						\
		})
			
		clock_gettime( CLOCK_REALTIME, &ts_before );
		for(x=0; x<10; x++) {
			lpm_test1_quiet( "004922191391721" );	/* DE: 0049221 prefix */
			lpm_test1_quiet( "00492238922733" );	/* DE: 00492238 prefix */
			lpm_test1_quiet( "004922251234567" );	/* DE: 0049 generic prefix */
			lpm_test1_quiet( "0015234" );		/* US: 001 generic prefix */
			lpm_test1_quiet( "00150089" );		/* US: 001500 prefix */
		}
		clock_gettime( CLOCK_REALTIME, &ts_after );

		tdiff = (float)ts_diff( ts_after, ts_before ) / 1000;

		/* output some stats */
		printf( "\n   --- Stats ---\n" );
		printf( "   Number of lookups performed: %d keys\n", num_routes); 
		printf( "   Average route key length   : %.2f chars\n", ((float)sum_length/(float)num_routes));
		printf( "   Time for lookups           : %-5.3f usecs\n", tdiff );
		printf( "   Average time per lookup    : %-5.3f usecs\n", (tdiff/(float)num_routes) );


		/*
		 * 13.3: valid results
		 */
		printf( "\n  * Part 3: LPM Route lookup tests, validity\n" );
		num_routes=0;
		failed_routes=0;

		#define lpm_test2(_key, _res) ({						\
			char *_x = _key;							\
			val = trie_get_partial( &lpm_trie, _x, strlen(_x) );			\
			if( !val )								\
				printf( "   failed to find route for key %-25s\n", _x );	\
			else {									\
				printf( "   key %-25s -> %-25s : ", _x, _res );			\
				if( !strcmp( val, _res ) )					\
					printf( "OK\n" );					\
				else {								\
					printf( "FAILED [%s]\n", val );				\
					failed_routes++;					\
				}								\
			}									\
			num_routes++;								\
		})
			
		lpm_test2( "004922191391721", "0049221" );	/* DE: 0049221 prefix */
		lpm_test2( "00492238922733",  "00492238" );	/* DE: 00492238 prefix */
		lpm_test2( "004922251234567", "0049" );		/* DE: 0049 generic prefix */

		printf( "   %d of %d routes are wrong\n", failed_routes, num_routes );

//		trie_destroy( &lpm_trie );


		/*
		 * 13.3b: valid results after removing some elements
		 */
		printf( "\n  * Part 3b: LPM Route lookup tests, validity #2\n" );
		num_routes=0;
		failed_routes=0;

		printf( "Removing prefix %s\n", prefixes[1].prefix );
		x = trie_del( &lpm_trie, prefixes[1].prefix, strlen(prefixes[1].prefix) );
		if( x < 0 )
			printf( "Error removing key: %d\n", x );

		printf( "Removing prefix %s\n", prefixes[3].prefix );
		x = trie_del( &lpm_trie, prefixes[3].prefix, strlen(prefixes[3].prefix) );
		if( x < 0 )
			printf( "Error removing key: %d\n", x );

		#define lpm_test2(_key, _res) ({						\
			char *_x = _key;							\
			val = trie_get_partial( &lpm_trie, _x, strlen(_x) );			\
			if( !val )								\
				printf( "   failed to find route for key %-25s\n", _x );	\
			else {									\
				printf( "   key %-25s -> %-25s : ", _x, _res );			\
				if( !strcmp( val, _res ) )					\
					printf( "OK\n" );					\
				else {								\
					printf( "FAILED [%s]\n", val );				\
					failed_routes++;					\
				}								\
			}									\
			num_routes++;								\
		})
			
		lpm_test2( "004922191391721", "0049" );		/* DE: 0049221 prefix */
		lpm_test2( "00492238922733",  "0049" );		/* DE: 00492238 prefix */
		lpm_test2( "004922251234567", "0049" );		/* DE: 0049 generic prefix */

		printf( "   %d of %d routes are wrong\n", failed_routes, num_routes );

		trie_destroy( &lpm_trie );


#ifdef BENCHMARKS
		/*
		 * 13.4: speed
		 */
		printf( "\n  * Part 4: LPM Route lookup tests (no output on successfull match)\n" );
		num_routes=0;
		sum_length=0;
		struct trie_node lpm2_trie;
		char buf[21];
		long len = 0;
		int y = 0, num_def_routes=0;

#define NUM_ROUTES	500000

		trie_init( &lpm2_trie );

		// default route
		lpm2_trie.flags |= TRIE_NODE_TERMINAL;
		lpm2_trie.ptr = &val2;

		// add routes
		for( x=0; x<NUM_ROUTES; x++ ) {
			len = random() % 20;

			for( y=0; y<len; y++ )
				buf[y] = 48 + (random() % 10);

			buf[len] = '\0';
			trie_add( &lpm2_trie, buf, strlen(buf), &val1 );
		}
		printf( "   %d prefixes added\n", x );

		#define lpm_test3_quiet(_key) ({						\
			char *_x = _key;							\
			int *_val;								\
			_val = trie_get_partial( &lpm2_trie, _x, strlen(_x) );			\
			if( !_val )								\
				printf( "   failed to find route for key %-25s\n", _x );	\
			else if( *_val == val2 )						\
				num_def_routes++;						\
			num_routes++;								\
			sum_length+= strlen(_x);						\
		})

			
		clock_gettime( CLOCK_REALTIME, &ts_before );
		for(x=0; x<NUM_ITER; x++) {
			len = 10 + random() % 10;

			for( y=0; y<len; y++ )
				buf[y] = 48 + (random() % 10);

			buf[len] = '\0';
			lpm_test3_quiet( buf );	/* DE: 0049 generic prefix */
		}
		clock_gettime( CLOCK_REALTIME, &ts_after );

		tdiff = (float)ts_diff( ts_after, ts_before ) / 1000;

		/* output some stats */
		printf( "\n   --- Stats ---\n" );
		printf( "   Number of lookups performed: %d keys\n", num_routes); 
		printf( "   Number of default routes   : %d keys\n", num_def_routes);
		printf( "   Average route key length   : %.2f chars\n", ((float)sum_length/(float)num_routes));
		printf( "   Time for lookups           : %-5.3f usecs\n", tdiff );
		printf( "   Average time per lookup    : %-5.3f usecs\n", (tdiff/(float)num_routes) );


		/*
		 * 14: speed of destroying
		 */
		printf( "\n*** Test 14: Time for destroying a (max.) %d key trie\n", NUM_ROUTES );

		clock_gettime( CLOCK_REALTIME, &ts_before );
		trie_destroy( &lpm2_trie );
		clock_gettime( CLOCK_REALTIME, &ts_after );

		tdiff = (float)ts_diff( ts_after, ts_before ) / 1000;

		/* output some stats */
		printf( "   Time for destruction           : %-5.3f usecs\n", tdiff );

		/* sanity check */
		for( x=0; x<TRIE_NUM_SLOTS; x++ ) {
			if( lpm2_trie.children[x] )
				break;
		}
		if( x < TRIE_NUM_SLOTS )
			printf( "!!! Root node still has children\n" );
#endif
	}

	fflush(stderr);

	printf( "\n%d of %d tests failed\n", num_failed, num_tests );
	return 0;
}

