#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>

//#define TRIE_USE_BITMAP
#define TRIE_PATHCOMP

//#define TRIE_DEBUG

#ifdef TRIE_USE_BITMAP
#define TRIE_USE_BITMAP_LOOKUP
#include "bitmap.h"
#endif
#include "trie.h"

#if defined(__GNUC__)
#define unlikely(cond) __builtin_expect(!!(cond), 0)
#define   likely(cond) __builtin_expect(!!(cond), 1)
#define	prefetch(x)    __builtin_prefetch(x)
#else
#define unlikely(cond)	(cond)
#define   likely(cond)	(cond)
#define prefetch(x)
#endif

#ifdef TRIE_PATHCOMP
# if defined(__WORDSIZE)
#  if __WORDSIZE == 64
#   define BYTES_PER_LONG	8
#   define PATH_SIZE_SHIFT	16
#   define PATH_SIZE_MASK	(0xffUL << PATH_SIZE_SHIFT)
#   define PATH_ELEMENTS	(__WORDSIZE / TRIE_SLOT_BITS)
#  else
#   define BYTES_PER_LONG	4
#   define PATH_SIZE_SHIFT	16
#   define PATH_SIZE_MASK	(0xffUL << PATH_SIZE_SHIFT)
#   define PATH_ELEMENTS	(__WORDSIZE / TRIE_SLOT_BITS)
#  endif
# else
#  error __WORDSIZE not available
# endif
#define PATH_SLOT_MASK		(~((~0UL) << TRIE_SLOT_BITS))
#define PATH_POS_SHIFT(p)	((p) * TRIE_SLOT_BITS)

#define PATH_GETSIZE(x)		(((x)->flags & PATH_SIZE_MASK) >> PATH_SIZE_SHIFT)
#define PATH_SETSIZE(x, v)	(x)->flags = ((x)->flags & ~PATH_SIZE_MASK) | (v << PATH_SIZE_SHIFT)

#define PATH_GETPOS(x, p)	(((x)->path >> PATH_POS_SHIFT(p)) & PATH_SLOT_MASK)
#define PATH_SETPOS(x, p, v)	(x)->path = ((x)->path & (~(PATH_SLOT_MASK << PATH_POS_SHIFT(p)))) | ((v & PATH_SLOT_MASK) << PATH_POS_SHIFT(p))

#define PATH_COPY_MASK(n)	(~((~0UL) << PATH_POS_SHIFT(n)))
#define PATH_COPY(to, from, to_off, from_off, len) ({			\
		unsigned long __tmp = 0;				\
		if( !from_off )						\
			__tmp = (from)->path & PATH_COPY_MASK(len));	\
		else {							\
			__tmp = (from)->path & (PATH_COPY_MASK(len) << PATH_POS_SHIFT(from_off)); \
			__tmp >>= PATH_POS_SHIFT(from_off);		\
		}							\
		(to)->path = ((to)->path & PATH_COPY_MASK(to_off)) | (__tmp << PATH_POS_SHIFT(to_off)); \
	})

#define PATH_FLAGS_NOCOPY	(PATH_SIZE_MASK | TRIE_NODE_ROOT | TRIE_NODE_TERMINAL)
#endif

#ifdef TRIE_DEBUG
#define DEBUGP(x...)	printf(## x)
#else
#define DEBUGP(x...)
#endif

/*
 * TODO:
 *	- make node.children dynamic to save space (bitmaps)
 *	- path compression (WIP, partially done)
 */
static int trie_node_child_count( const struct trie_node *node );

#if TRIE_NUM_SLOTS == 10
static inline char ascii2bcd(const char val)
{
	if(val >= 30 && val < 40)
		return val - 30;

	return 0;
}
#else
#define ascii2bcd(x)	(x)
#endif

void trie_dump_node(const struct trie_node *n, const char *key, int keylen)
{
	int x;

	printf( "--- Trie node %p ---\n", n );
	printf( "Parent node %p\n", n->parent );
	printf( "Leaf? %s\n", n->flags & TRIE_NODE_LEAF ? "yes" : "no" );
	printf( "  Children: %d\n", trie_node_child_count(n) );
	printf( "Root? %s\n", n->flags & TRIE_NODE_ROOT ? "yes" : "no" );
	printf( "Terminal? %s\n", n->flags & TRIE_NODE_TERMINAL ? "yes" : "no" );
	if( n->flags & TRIE_NODE_TERMINAL ) {
		printf( "  Data ptr: %p\n", n->ptr );
	}
	printf( "Compressed path? %s\n", n->flags & TRIE_NODE_COMPRESSED ? "yes" : "no" );
	if(n->flags & TRIE_NODE_COMPRESSED) {
		printf( "  Length of path: %lu\n", PATH_GETSIZE(n));
		printf( "  Path: %lx\n", n->path);
	}
	if( key ) {
		printf( "Current key: " );
		for( x=0; x<keylen; x++ ) {
			printf( "%c", key[x] );
		}
	}
	printf( "\n--------------------\n" );
}

#ifdef TRIE_PATHCOMP
/*
 * Pack prefix into node
 */
int trie_compress_prefix( struct trie_node *node, const char *prefix, int size )
{
	int off, pos = 0;

	if( size <= 0 )
		return 0;

	for( off = PATH_GETSIZE(node); off < PATH_ELEMENTS && pos < size; off++, pos++ ) {
		PATH_SETPOS( node, off, ascii2bcd(prefix[pos]) );
	}

	if( pos ) {
		PATH_SETSIZE( node, off );
		node->flags |= TRIE_NODE_COMPRESSED;
	}
	return pos;
}

struct trie_node *trie_split_prefix( struct trie_node *node, const char node_key, int where )
{
	struct trie_node *tmp;

	if( !node || !(node->flags & TRIE_NODE_COMPRESSED) )
		return NULL;

	if( where < 0 || where >= PATH_ELEMENTS )
		return NULL;

	DEBUGP( "\t++   Splitting node %p (%c, pfxlen: %d, pfx: %#lx) at offset %d, pfxlen %d\n",
		node, node_key, PATH_GETSIZE(node), node->path, where, PATH_GETSIZE(node) );

	/* temporary node that will replace cur */
	tmp = malloc( sizeof(struct trie_node) );
	if( unlikely(!tmp) )
		return NULL;

	memset( tmp, 0, sizeof(struct trie_node) );

	tmp->flags = node->flags & (~PATH_FLAGS_NOCOPY);		/* flags */

	if( where ) {
		tmp->path   = node->path & PATH_COPY_MASK(where);	/* copy common part of path */
		tmp->flags |= TRIE_NODE_COMPRESSED;
	}

	tmp->children[PATH_GETPOS(node, where)] = node;		/* make cur a child of the new node */
	tmp->parent = node->parent;				/* new parent */

	node->path >>= PATH_POS_SHIFT( where + 1 );		/* remove digits */
	PATH_SETSIZE( node, (PATH_GETSIZE(node)-(where+1)) );	/* update length */
	node->parent->children[(int)node_key] = tmp;		/* update parent */
	node->parent = tmp;					/* new parent */

	if( PATH_GETSIZE(node) == 0 )				/* remove compressed flag */
		node->flags &= ~TRIE_NODE_COMPRESSED;

	PATH_SETSIZE( tmp, where );

	DEBUGP( "\t++ Node %p split (pfxlen: %d, pfx: %#lx), new node %p (pfxlen: %d, pfx: %#lx)\n",
		node, PATH_GETSIZE(node), node->path, tmp, PATH_GETSIZE(tmp), tmp->path );

	return tmp;
}

struct trie_node *trie_merge_prefix( struct trie_node *node, struct trie_node *into, const char node_key )
{
	int into_pfxlen = 0, node_pfxlen = 0, space = 0;

	if( !node || !into )
		return NULL;

	/* Target is terminal or root node? No worky (for full and partial merging!) */
	if( into->flags & (TRIE_NODE_TERMINAL | TRIE_NODE_ROOT) )
		return NULL;

	into_pfxlen = PATH_GETSIZE(into);
	node_pfxlen = PATH_GETSIZE(node);
	space  = PATH_ELEMENTS - into_pfxlen;

	/* no free space left in prefix, merging not possible */
	if( space <= 0 )
		return NULL;

	/* Merge impossible if target node has multiple children OR target node is not it's parent */
	if( trie_node_child_count(into) > 1 || node->parent != into )
		return NULL;

	/* Is the resulting prefix longer than the space available? */
	if( (node_pfxlen + 1) > space ) {
		struct trie_node *tmp;

		/*
		 * Full merge impossible, try partial merge (e.g. move some of the prefix into the into node)
		 */

		/* Partial merge not possible if target node has a value */
		if( into->flags & TRIE_NODE_TERMINAL )
			return NULL;

		/* Space in target node available? (min. 2 elements) */
		if( space >= 2 )
			return NULL;

		/* Slot in target node available? */
		tmp = into->children[(int)node_key];
		if( tmp && tmp != node )
			return NULL;

		DEBUGP( "\t**   Partially merging prefix of %p (child: %d, pfxlen: %d, pfx: %#lx) into prefix of %p (as child: %d, pfxlen: %d, pfx: %#lx)\n",
			node, (int)node_key, node_pfxlen, node->path, into, (int)PATH_GETPOS(node, (space - 1)), into_pfxlen, into->path ); 

		if( space > 1 ) {
			into->path = (into->path & PATH_COPY_MASK(into_pfxlen)) | (((unsigned long)node_key) << PATH_POS_SHIFT(into_pfxlen)) | (node->path << PATH_POS_SHIFT(into_pfxlen + 1));
		} else {
			into->path = (into->path & PATH_COPY_MASK(into_pfxlen)) | (((unsigned long)node_key) << PATH_POS_SHIFT(into_pfxlen));
		}

		into->children[(int)node_key] = NULL;
		into->children[PATH_GETPOS(node, (space - 1))] = node;
		PATH_SETSIZE(into, PATH_ELEMENTS);

		node->path >>= PATH_POS_SHIFT(space);
		PATH_SETSIZE(node, (node_pfxlen - space));

		DEBUGP( "\t**     Merged node %p (pfxlen: %d, pfx: %#lx) into %p (pfxlen: %d, pfx: %#lx)\n", node, PATH_GETSIZE(node), node->path, into, PATH_GETSIZE(into), into->path );

		/* done */
	} else {
		/*
		 * Full merge
		 */
		DEBUGP( "\t**   Fully merging %p (child %d, pfxlen: %d, pfx: %#lx) into %p (pfxlen: %d, pfx: %#lx)\n",
			node, (int)node_key, node_pfxlen, node->path, into, into_pfxlen, into->path ); 

		if( !into_pfxlen && !node_pfxlen ) {
			into->path = node_key;
			PATH_SETSIZE( into, 1 );
		}
		else if( !node_pfxlen ) {
			into->path = (into->path & PATH_COPY_MASK(into_pfxlen)) | (((unsigned long)node_key) << PATH_POS_SHIFT(into_pfxlen));
			PATH_SETSIZE( into, (into_pfxlen + 1) );
		}
		else if( !into_pfxlen ) {
			into->path = ((node->path & PATH_COPY_MASK(node_pfxlen)) << PATH_POS_SHIFT(1)) | ((unsigned long)node_key);
			PATH_SETSIZE( into, (node_pfxlen + 1) );
		}
		else {
			into->path = (into->path & PATH_COPY_MASK(into_pfxlen)) | (((unsigned long)node_key) << PATH_POS_SHIFT(into_pfxlen)) | (node->path << PATH_POS_SHIFT(into_pfxlen + 1));
			PATH_SETSIZE( into, (into_pfxlen + node_pfxlen + 1) );
		}

		into->children[(int)node_key] = NULL;
		into->flags |= TRIE_NODE_COMPRESSED;

		if( node->flags & TRIE_NODE_LEAF ) {
			memcpy( into->children, node->children, sizeof(struct trie_node *) * TRIE_NUM_SLOTS );
			into->flags |= TRIE_NODE_LEAF;

			/*
			 * meh, brainfart, the old code avoided this by merging "into"
			 * into "node" and just updating the parent (this would save us
			 * from doing the memcpy too, at the expense of having to rewrite
			 * or split this function)
			 */
			for( int x = 0; x < TRIE_NUM_SLOTS; x++ ) {
				struct trie_node *tmp = into->children[x];
				if( tmp )
					tmp->parent = into;
			}
		}

		if( node->flags & TRIE_NODE_TERMINAL ) {
			into->ptr = node->ptr;
			into->flags |= TRIE_NODE_TERMINAL;
		}

		DEBUGP( "\t**     Resulting node %p (pfxlen: %d, pfx: %#lx)\n", into, PATH_GETSIZE(into), into->path );

		free( node );
	}

	return into;
}
#endif


int trie_init( struct trie_node *n )
{
	if( !n )
		return -1;

	memset( n, 0, sizeof(struct trie_node) );

	n->flags = TRIE_NODE_ROOT | TRIE_NODE_LEAF;

#ifdef TRIE_USE_BITMAP_LOOKUP
	bitmap_clear_all( n->map, TRIE_NUM_SLOTS );
#endif
	return 0;
}


void *trie_get( const struct trie_node *n, const char *key, const int keylen )
{
	struct trie_node *cur = (struct trie_node *)n;
	int i;
#ifdef TRIE_PATHCOMP
	int offset = 0;
#endif

	if( !n || !key || !keylen )
		return NULL;

	if( keylen > TRIE_MAX_DEPTH )
		return NULL;

	for( i=0; i<keylen; i++ ) {

#ifdef TRIE_PATHCOMP
		if( cur->flags & TRIE_NODE_COMPRESSED ) {
			int psize = PATH_GETSIZE(cur);

			if( offset < psize ) {
				if( ascii2bcd(key[i]) != PATH_GETPOS(cur, offset) )
					return NULL;

				offset++;
				continue;
			} else 
				offset = 0;
		}
#endif
#ifdef TRIE_USE_BITMAP_LOOKUP
		if( !bitmap_is_set( cur->map, TRIE_NUM_SLOTS, (int)key[i] ) )
			return NULL;
#endif
		cur = cur->children[(int)key[i]];
		if( unlikely( !cur ) )
			return NULL;

#ifdef TRIE_USE_BITMAP_LOOKUP
		prefetch(cur->map);
#endif
	}
#ifdef TRIE_PATHCOMP
	if( offset == PATH_GETSIZE(cur) && (cur->flags & TRIE_NODE_TERMINAL) )
		return cur->ptr;
#else
	if( cur->flags & TRIE_NODE_TERMINAL )
		return cur->ptr;
#endif
	return NULL;
}

void *trie_get_partial( const struct trie_node *n, const char *key, const int keylen )
{
	struct trie_node *cur = (struct trie_node *)n;
	struct trie_node *last_term_node = NULL;
	int i;
#ifdef TRIE_PATHCOMP
	int offset = 0;
#endif

	if( !n || !key || !keylen )
		return NULL;

	for( i=0; i<keylen; i++ ) {

#ifdef TRIE_PATHCOMP
		if( cur->flags & TRIE_NODE_COMPRESSED ) {
			int psize = PATH_GETSIZE(cur);

			if( offset < psize ) {
				if( ascii2bcd(key[i]) != PATH_GETPOS(cur, offset) ) {
					DEBUGP( "node %p (parent %p) prefix: %#lx, len: %d does no match fully!\n", cur, cur->parent, cur->path, PATH_GETSIZE(cur) );
					cur = NULL;
					goto out;
				}
				offset++;
				continue;
			} else {
				offset = 0;

				DEBUGP( "node %p (parent %p) prefix: %#lx, len: %d, flags [%c%c] matched fully!\n", cur, cur->parent, cur->path, PATH_GETSIZE(cur),
					(cur->flags & TRIE_NODE_TERMINAL) ? 'T' : ' ', (cur->flags & TRIE_NODE_LEAF) ? 'L' : ' ' );

				if( cur->flags & TRIE_NODE_TERMINAL )
					last_term_node = cur;
			}
		} else {
			DEBUGP( "node %p (parent %p) prefix: %#lx, len: %d, flags [%c%c] not compressed\n", cur, cur->parent, cur->path, PATH_GETSIZE(cur),
				(cur->flags & TRIE_NODE_TERMINAL) ? 'T' : ' ', (cur->flags & TRIE_NODE_LEAF) ? 'L' : ' ' );

			if( cur->flags & TRIE_NODE_TERMINAL )
				last_term_node = cur;
		}
#endif

#ifdef TRIE_USE_BITMAP_LOOKUP
		if( !bitmap_is_set( cur->map, TRIE_NUM_SLOTS, (int)key[i] ) )
			break;
#endif

		cur = cur->children[(int)key[i]];
		if( unlikely( !cur ) )
			break;

#ifndef TRIE_PATHCOMP
		if( cur->flags & TRIE_NODE_TERMINAL )
			last_term_node = cur;
#endif
#ifdef TRIE_USE_BITMAP_LOOKUP
		prefetch(cur->map);
#endif
	}
out:
	if( cur && ( cur->flags & TRIE_NODE_TERMINAL ) )
		return cur->ptr;

	if( last_term_node )
		return last_term_node->ptr;

	return NULL;
}


int trie_has_key( const struct trie_node *n, const char *key, const int keylen )
{
	return trie_get( n, key, keylen ) ? 1 : 0;
}

int trie_has_key_partial( const struct trie_node *n, const char *key, const int keylen )
{
	return trie_get_partial( n, key, keylen ) ? 1 : 0;
}


int trie_add( struct trie_node *n, const char *key, const int keylen, const void *ptr )
{
	struct trie_node *cur = n, *next = NULL;
	int i = 0, val = 0;
#ifdef TRIE_PATHCOMP
	int skip_path = 0;
#endif

	if( !n || !key || !keylen || !ptr )
		return -EINVAL;

	if( keylen > TRIE_MAX_DEPTH )
		return -EINVAL;

	for( i=0; i<keylen; i++ ) {

#ifdef TRIE_PATHCOMP
		if( cur->flags & TRIE_NODE_COMPRESSED && !skip_path ) {
			int psize = PATH_GETSIZE(cur);
			int off = 0, split_node = 0;

			/*
			 * Walk the compressed path that is recorded in this node
			 */
			while(i < keylen && off < psize) {
				if( ascii2bcd(key[i+off]) != PATH_GETPOS(cur, off) ) {
					split_node = 1;		/* Path needs to be split at this position */
					break;
				}
				off++;
			}

			/*
			 * Complex case, existing path has to be split
			 */
			if( split_node ) {
				next = trie_split_prefix( cur, key[i-1], off );
				if( !next )
					return -1;

				cur = next;
			}
			i += off;
		}
		skip_path = 0;
#endif
		val = key[i];

#ifdef TRIE_USE_BITMAP_LOOKUP
		if( !bitmap_is_set( cur->map, TRIE_NUM_SLOTS, val ) ) {
#else
		next = cur->children[val];
		if( !next ) {
#endif
			int ret;

			DEBUGP( "\t++ Adding new leaf for node %p [%d of %d: (%#x)]\n", cur, i, keylen, val );

			next = malloc( sizeof(struct trie_node) );
			if( unlikely(!next) )
				return -1;

			memset( next, 0, sizeof(struct trie_node) );

			if( !(cur->flags & TRIE_NODE_LEAF) )
				cur->flags |= TRIE_NODE_LEAF;
#ifdef TRIE_PATHCOMP
			/*
			 * New leaf / after path split, more than one element left in key
			 */
			ret = trie_compress_prefix( next, &key[i+1], keylen - (i + 1) );
			if( ret ) {
				i += ret;
				skip_path = 1;
			}
#endif
			cur->children[val] = next;
			next->parent = cur;

#ifdef TRIE_USE_BITMAP
			DEBUGP( "\t++ Marking slot %d used in node %p (pointing to %p) @level %d of %d\n", val, cur, next, i, keylen );
			bitmap_set( cur->map, TRIE_NUM_SLOTS, val );
#endif
		}
#ifdef TRIE_USE_BITMAP_LOOKUP
		else
			cur = cur->children[val];
#else
		cur = next;
#endif
	}
	if( unlikely(cur->flags & TRIE_NODE_TERMINAL) ) {
		/*
		 * collision(!)
		 */
		DEBUGP( "\t--- collision detected (%p: %d/%d)?\n", cur, i, keylen );
		return -EEXIST;
	}

	DEBUGP( "\t+++ Flipping terminal flag on node %p [(%#x)]\n", cur, val );

	cur->ptr = (void *)ptr;
	cur->flags |= TRIE_NODE_TERMINAL;

	return 0;
}


static inline int trie_node_unused( const struct trie_node *node )
{
	if( node->flags & TRIE_NODE_TERMINAL )
		return 0;

#ifdef TRIE_USE_BITMAP
	if( (node->flags & TRIE_NODE_LEAF) && !bitmap_is_empty( node->map, TRIE_NUM_SLOTS ) )
		return 0;
#else
	{
		struct trie_node **ptr = (void *)node->children;

		for(int x = 0; x < TRIE_NUM_SLOTS; x++, ptr++ ) {
			if( *ptr )
				return 0;
		}
	}
#endif
	return 1;
}

static int trie_node_child_count( const struct trie_node *node )
{
	struct trie_node **ptr = (void *)node->children;
	int count = 0;

	for(int x = 0;  x < TRIE_NUM_SLOTS; x++, ptr++ ) {
		if( *ptr )
			count++;
	}
	return count;
}

static inline void *trie_node_child_get( const struct trie_node *node, int *offset )
{
	struct trie_node **ptr = (void *)node->children + *offset;

	for(int x = *offset; x < TRIE_NUM_SLOTS; x++, ptr++ ) {
		if( *ptr ) {
			*offset = x;
			return *ptr;
		}
	}
	return NULL;
}


/*
 * not perfect, slow...
 */
int trie_del( struct trie_node *n, const char *key, const int keylen )
{
	struct trie_node *cur = n, *next;
	int i;
#ifdef TRIE_PATHCOMP
	int offset = 0;
#endif

	if( !n || !key || !keylen )
		return -EINVAL;

	if( keylen > TRIE_MAX_DEPTH )
		return -EINVAL;

	for( i=0; i<keylen; i++ ) {
#ifdef TRIE_PATHCOMP
		if( cur->flags & TRIE_NODE_COMPRESSED ) {
			int psize = PATH_GETSIZE(cur);

			if( offset < psize ) {
				if( ascii2bcd(key[i]) != PATH_GETPOS(cur, offset) ) {
					/* prefix mismatch, key not in trie */
					return -1;
				}
				offset++;
				continue;
			} else {
				offset = 0;
			}
		}
#endif
		cur = cur->children[(int)key[i]];
		if( unlikely(!cur) ) {
			DEBUGP( "%p has no child key[%d]->(%c)\n", cur, i, key[i] );
			return -EEXIST;
		}
	}
	if( unlikely(!(cur->flags & TRIE_NODE_TERMINAL)) ) {
		/*
		 * not a terminal node, key not in trie...
		 */
		DEBUGP( "%p is not a terminal node\n", cur );
		return -EEXIST;
	}

	cur->flags &= ~TRIE_NODE_TERMINAL;	/* found key, drop terminal flag on node */

	DEBUGP( "\t++ Starting at node %p [%d [%#x] = %c]\n", cur, i-1, key[i-1], key[i-1] );

	/*
	 * walk trie backwards removing unused leaf nodes,
	 * abort as soon as we find a node that's still in use
	 */
	for( i--; i > 0; i-- ) {

		next = cur->parent;

#ifdef TRIE_PATHCOMP
		/*
		 * Strip n = prefixlen components from key
		 */
		if( cur->flags & TRIE_NODE_COMPRESSED ) {
			i -= PATH_GETSIZE(cur);
		}
#endif
		if( !(cur->flags & (TRIE_NODE_LEAF | TRIE_NODE_TERMINAL)) ) {
			int num;

			/*
			 * Remove
			 */
			DEBUGP( "\t++ Removing unused leaf node %p [(%#x)]\n", cur, key[i] );
#ifdef TRIE_USE_BITMAP
			bitmap_clear( next->map, TRIE_NUM_SLOTS, (int)key[i] );
#endif
			next->children[(int)key[i]] = NULL;
			free( cur );

			num = trie_node_child_count(next);
			if( !num ) {
				/* Drop leaf flag */
				next->flags &= ~TRIE_NODE_LEAF;
			}
#ifdef TRIE_PATHCOMP
			if( num == 1 && !(next->flags & TRIE_NODE_ROOT) ) {
				struct trie_node *tmp = NULL;
				int offset = 0;

				/* Only one child left, merge into parent */
				DEBUGP( "\t** Parent node %p has only 1 child, attempting to merge\n", next );

				tmp = trie_node_child_get( next, &offset );
				if( !tmp )
					return -1;

				trie_merge_prefix( tmp, next, offset );
			}
#endif
		}
#ifdef TRIE_PATHCOMP
		else if( !(cur->flags & TRIE_NODE_TERMINAL) && (trie_node_child_count(cur) == 1) ) {
			struct trie_node *tmp = NULL;
			int offset = 0;

			/* only one child node, merge */
			DEBUGP( "\t** Node %p has only 1 child, attempting to merge child into parent\n", cur );

			tmp = trie_node_child_get( cur, &offset );
			if( !tmp )
				return -1;

			trie_merge_prefix( tmp, cur, offset );
		}
#endif
		else {
			DEBUGP( "\t-- Node %p still in use, not removing\n", cur );
//			break;
		}

		cur = next;
	}
	return 0;
}

/*
 * probably fscking slow...
 */
int trie_dump_destroy( struct trie_node *n, void (* dump_cb)(const char *key, const int keylen, const void *val, const void *priv), const void *priv )
{
	struct trie_node *cur = n, *prev = NULL;
	int level = 0, x;
	char key[TRIE_MAX_DEPTH] = { 0 };

	if( !n )
		return -EINVAL;

	do {
		/*
		 * empty slot, go back to parent
		 */
		if( unlikely(!cur) ) {
			DEBUGP( "\t** Going down one level (%d -> %d) node %p parent %p\n", level, level - 1, cur, prev );

			cur = prev;
			level--;
		}

		/*
		 * end node? clear terminal flag
		 */
		if( cur->flags & TRIE_NODE_TERMINAL ) {
			if( dump_cb )
				dump_cb( key, level, cur->ptr, priv );

			cur->flags &= ~TRIE_NODE_TERMINAL;
			cur->ptr = NULL;
		}

		/*
		 * has children?
		 */
		if( cur->flags & TRIE_NODE_LEAF ) {
			/*
			 * Todo: add bitmap_first_bit function to get the first set bit in map
			 *       for faster lookup
			 */
#ifdef TRIE_USE_BITMAP
			x = bitmap_first_bit( cur->map, TRIE_NUM_SLOTS ) - 1;
			if( x >= 0 ) {
				/*
				 * yes, go to next level
				 */
				cur = cur->children[x];
				key[level++] = x;
				continue;
			}
#else
			for( x = 0; x < TRIE_NUM_SLOTS; x++ ) {
				if( cur->children[x] )
					break;
			}
			if( x < TRIE_NUM_SLOTS ) {
				/*
				 * yes, go to next level
				 */
				DEBUGP( "\t** Going up one level (%d -> %d) node %p child %d (pfxlen: %d, pfx: %#lx)\n", level, level + 1,
					cur, x, PATH_GETSIZE((struct trie_node *)cur->children[x]), ((struct trie_node *)cur->children[x])->path );

				cur = cur->children[x];
				key[level++] = x;
#ifdef TRIE_PATHCOMP
				/* go prefix up */
				if( cur && (cur->flags & TRIE_NODE_COMPRESSED) ) {
					for( x = 0; x < PATH_GETSIZE(cur); x++, level++ )
						key[level] = PATH_GETPOS( cur, x );
				}
#endif
				continue;
			}
#endif
			/*
			 * no children, drop leaf flag
			 */
			cur->flags &= ~TRIE_NODE_LEAF;
			DEBUGP( "\t** Node has no children, dropping leaf flag\n" );
		}

#ifdef TRIE_PATHCOMP
		/* go pfx down */
		if( cur->flags & TRIE_NODE_COMPRESSED ) {
			int pfxlen = PATH_GETSIZE( cur );

			DEBUGP( "\t** Going down prefix of length %d (level %d -> %d)\n", pfxlen, level, level - pfxlen );

			memset( &key[level - pfxlen], 0, pfxlen );
			level -= pfxlen;
		}
#endif

		/*
		 * can't delete root node
		 */
		if( !level ) {
			break;
		}

		/*
		 * completely unused, delete node
		 */
		prev = cur->parent;
		DEBUGP( "\t** Freeing child node %p of parent %p and going down one level (%d -> %d)\n", cur, prev, level, level - 1 );
		prev->children[(int)key[--level]] = NULL;
#ifdef TRIE_USE_BITMAP
		bitmap_clear( prev->map, TRIE_NUM_SLOTS, key[level] );
#endif
		free( cur );
		cur = prev;

	} while( 1 );

	return 0;
}


int trie_destroy( struct trie_node *n )
{
	return trie_dump_destroy( n, NULL, NULL );
}


int trie_traverse( const struct trie_node *n,
		   void (* trav_cb)(const struct trie_node *n, const char *key, const int keylen, const void *value, void *data),
		   void *cb_priv,
		   const int flags )
{
	struct trie_node *cur = (void *)n, *prev = NULL;
	int level = 0, x;
	char key[TRIE_MAX_DEPTH] = { 0 };

	if( !n || !trav_cb || !flags )
		return -EINVAL;

	do {
		/*
		 * empty slot, go back to parent
		 */
		if( unlikely(!cur) ) {
			cur = prev;
			level--;
		}

		/*
		 * invoke callback (only once)
		 */
		if( !key[level] &&
		    (((cur->flags & TRIE_NODE_TERMINAL) && (flags & TRIE_TRAVERSE_TERM)) ||
		     ((cur->flags & TRIE_NODE_LEAF)     && (flags & TRIE_TRAVERSE_LEAF))) )
			trav_cb( cur, key, level, cur->ptr, cb_priv );

		/*
		 * has children?
		 */
		if( cur->flags & TRIE_NODE_LEAF ) {
			/*
			 * Todo: add bitmap_first_bit function to get the first set bit in map
			 *       for faster lookup
			 */

			/*
			 * start where we left off...
			 */
			for( x=key[level]; x<TRIE_NUM_SLOTS; x++ ) {
				if( cur->children[x] )
					break;
			}
			if( x < TRIE_NUM_SLOTS ) {
				/*
				 * yes, go to child on next level
				 */
				cur = cur->children[x];
				key[level++] = x;
#ifdef TRIE_PATHCOMP
				/* go prefix up */
				if( cur && (cur->flags & TRIE_NODE_COMPRESSED) ) {
					for( x = 0; x < PATH_GETSIZE(cur); x++, level++ )
						key[level] = PATH_GETPOS( cur, x );
				}
#endif
				continue;
			} else {
				/*
				 * nope, reset level key slot
				 */
				key[level] = 0;
			}
		}

#ifdef TRIE_PATHCOMP
		/* go prefix down */
		if( cur->flags & TRIE_NODE_COMPRESSED ) {
			int pfxlen = PATH_GETSIZE( cur );

			memset( &key[level - pfxlen], 0, pfxlen );
			level -= pfxlen;
		}
#endif
		/*
		 * stop at root node
		 */
		if( !level )
			break;

		/*
		 * we're done on this level,
		 * go back up to next node in parent's list
		 */
		prev = cur->parent;
		cur = prev;
		key[--level]++;

	} while( 1 );

	return 0;
}
