#ifndef _TRIE_H
#define _TRIE_H

/*
 * Generic trie
 */
#ifndef TRIE_NUM_SLOTS
#define TRIE_NUM_SLOTS	256
#define TRIE_SLOT_BITS	8
//#define TRIE_NUM_SLOTS	10
//#define	TRIE_SLOT_BITS	4
#endif

#if defined(TRIE_PATHCOMP) && !defined(TRIE_SLOT_BITS)
#error TRIE_SLOT_BITS not set
#endif

#define TRIE_MAX_DEPTH  30


enum trie_node_flags {
	TRIE_NODE_ROOT = 1,
	TRIE_NODE_LEAF = 2,
	TRIE_NODE_TERMINAL = 4,
#ifdef TRIE_PATHCOMP
	TRIE_NODE_COMPRESSED = 8,
#endif
};

/*
 * Node sizes for a possible dynamic implementation (w/ path compression, w/o bitmaps)
 *
 * leaf:
 *   @32bit: 12 + 4 + n * 4   (= 1040 Bytes @256 Slots /  56 Bytes @10)
 *   @64bit: 24 + 8 + n * 8   (= 2080 Bytes @256 Slots / 112 Bytes @10)
 *
 * terminal:
 *   @32bit: 12 + 4
 *   @64bit: 24 + 8
 */
struct trie_node {
	unsigned long		flags;		/*!< Node flags */
#ifdef TRIE_PATHCOMP
	unsigned long		path;		/*!< Compressed path */
#endif
	struct trie_node	*parent;	/*!< Pointer to parent node */
	void			*ptr;		/*!< Data pointer */
#ifdef TRIE_USE_BITMAP
	BITMAP(map, TRIE_NUM_SLOTS);		/*!< In-use bitmap */
#endif
	void	*children[TRIE_NUM_SLOTS];	/*!< Leaf nodes */
};

/**
 * TRIE_HEAD
 * \param	name	Name of the variable
 * \brief	Declare and init trie head (root node)
 *
 * \author	Stefan Knoblich <s.knoblich@axsentis.de>
 */
#define TRIE_HEAD(name) 					\
	struct trie_node name = {				\
		.flags = TRIE_NODE_ROOT | TRIE_NODE_LEAF,	\
		.ptr   = NULL,					\
		.children = { 0 },				\
		.parent= NULL,					\
	}

/**
 * trie_init
 * \brief	Initialize trie
 * \param	n	pointer to trie base node
 * \return	0 on success, -1 on failure
 *
 * \author	Stefan Knoblich <s.knoblich@axsentis.de>
 */
int trie_init( struct trie_node *n );


/**
 * trie_has_key
 * \brief	Check if (exact) matching key is stored in trie
 * \param	n	pointer to trie base node
 * \param[in]	key	key to check
 * \param[in]	keylen	length of key
 * \return	-1 on error, 0 if key is not in the trie, 1 otherwise
 *
 * \author	Stefan Knoblich <s.knoblich@axsentis.de>
 */
int trie_has_key( const struct trie_node *n, const char *key, const int keylen );


/**
 * trie_has_key_exact
 * \brief	Alias to trie_has_key
 * \author	Stefan Knoblich <s.knoblich@axsentis.de>
 */
#define trie_has_key_exact	trie_has_key


/**
 * trie_has_key_partial
 * \brief	Check if (partial) matching key is stored in trie
 * \param	n	pointer to trie base node
 * \param[in]	key	key to check
 * \param[in]	keylen	length of key
 * \return	-1 on error, 0 if key is not in the trie, 1 otherwise
 *
 * \author	Stefan Knoblich <s.knoblich@axsentis.de>
 */
int trie_has_key_partial( const struct trie_node *n, const char *key, const int keylen );


/**
 * trie_get
 * \brief	Return value associated with (exact) matching key from trie
 * \param	n	pointer to trie base node
 * \param[in]	key	key to retrieve
 * \param[in]	keylen	length of key
 * \return	NULL on error, pointer stored otherwise
 *
 * \author	Stefan Knoblich <s.knoblich@axsentis.de>
 */
void *trie_get( const struct trie_node *n, const char *key, const int keylen );


/**
 * trie_get_exact
 * \brief	Alias to trie_get
 *
 * \author	Stefan Knoblich <s.knoblich@axsentis.de>
 */
#define trie_get_exact	trie_get


/**
 * trie_get_partial
 * \brief	Return value associated with (partial) matching key from trie
 * \param	n	pointer to trie base node
 * \param[in]	key	key to retrieve
 * \param[in]	keylen	length of key
 * \return	NULL on error, pointer stored otherwise
 *
 * \author	Stefan Knoblich <s.knoblich@axsentis.de>
 */
void *trie_get_partial( const struct trie_node *n, const char *key, const int keylen );


/**
 * trie_add
 * \brief	Add key (+value) to trie
 * \param	n	pointer to trie base node
 * \param[in]	key	key to add
 * \param[in]	keylen	length of key
 * \param[in]	ptr	pointer to value to store
 * \return	0 on success, -1 on failure
 *
 * \author	Stefan Knoblich <s.knoblich@axsentis.de>
 */
int trie_add( struct trie_node *n, const char *key, const int keylen, const void *ptr );


/**
 * trie_del
 * \brief	Delete key from trie
 * \param	n	pointer to trie base node
 * \param[in]	key	key to remove
 * \param[in]	keylen	length of key
 * \return	0 on success, -1 on failure
 *
 * \author	Stefan Knoblich <s.knoblich@axsentis.de>
 */
int trie_del( struct trie_node *n, const char *key, const int keylen );


/**
 * trie_destroy
 * \brief	Completely destroy trie
 * \param	n	pointer to trie root node
 * \return	0 on success; < 0 on failure
 *
 * \author	Stefan Knoblich <s.knoblich@axsentis.de>
 */
int trie_destroy( struct trie_node *n );


/**
 * trie_dump_destroy
 * \brief	Completely destroy trie and dump data via callback
 * \param	n	pointer to trie root node
 * \param	dump_cb	callback function for dumping the trie content
 * \param	priv	pointer to private data passed to dump_cb
 * \return	0 on success; < 0 on failure
 *
 * \author	Stefan Knoblich <s.knoblich@axsentis.de>
 */
int trie_dump_destroy( struct trie_node *n, void (* dump_cb)(const char *key, const int keylen, const void *val, const void *priv), const void *priv );


/**
 * \brief	Mode types for trie_traverse
 */
enum {
	TRIE_TRAVERSE_TERM = 1,	/*!< Invoke callback only on terminal nodes */
	TRIE_TRAVERSE_LEAF = 2, /*!< Invoke callback only on leaf nodes */
	TRIE_TRAVERSE_ALL  = 3  /*!< Invoke callback for all nodes */
};


/**
 * trie_traverse
 * \brief	Traverse trie and invoke callback for every terminal node
 * \param	n	trie root node pointer
 * \param	trav_cb	callback function pointer
 * \param	cb_priv	pointer to data to pass to callback function
 * \param	flags	one of TRIE_TRAVERSE_TERM / _LEAF or _ALL
 *
 * \author	Stefan Knoblich <s.knoblich@axsentis.de>
 */

int trie_traverse( const struct trie_node *n, void (* trav_cb)(const struct trie_node *n, const char *key, const int keylen, const void *value, void *priv), void *cb_priv, const int flags );


/**
 * trie_dump_node
 * \brief	Dump some information about the trie node to stdout
 * \note	Used for debugging
 * \param	n	trie node
 * \param	key	NULL or pointer to key array
 * \param	keylen	Length of key or 0
 *
 * \author	Stefan Knoblich <s.knoblich@axsentis.de>
 */
void trie_dump_node( const struct trie_node *n, const char *key, int keylen );

#endif

